"""Argparser module for hub build"""
from ..helper import add_arg_group
from ...enums import BuildTestLevel


def mixin_hub_build_parser(parser):
    """Add the arguments for hub build to the parser

    :param parser: the parser configure
    """
    gp = add_arg_group(parser, title='Build')
    gp.add_argument(
        'path',
        type=str,
        help='''
The content source to be shipped into a Docker image. It can one of the followings:
- a directory containing Dockerfile, manifest.yml, README.md, zero or more yaml config, zero or more Python file.
- a name of the Jina Hub executor given by `jina hub list --local-only`.
''',
    )
    gp.add_argument(
        '-f',
        '--file',
        type=str,
        default='Dockerfile',
        help='Name of the Dockerfile (Default is `path/Dockerfile`)',
    )
    gp.add_argument(
        '--pull',
        action='store_true',
        default=False,
        help='If set, downloads any updates to the FROM image in Dockerfiles',
    )
    gp.add_argument(
        '--push',
        action='store_true',
        default=False,
        help='If set, push the built image to the registry',
    )
    gp.add_argument(
        '--dry-run',
        action='store_true',
        default=False,
        help='If set, only check path and validity, no real building',
    )
    gp.add_argument(
        '--prune-images',
        action='store_true',
        default=False,
        help='If set, prune unused images after building, this often saves disk space',
    )
    gp.add_argument(
        '--raise-error',
        action='store_true',
        default=False,
        help='If set, raise any error and exit with code 1',
    )
    gp.add_argument(
        '--test-uses',
        action='store_true',
        default=False,
        help='If set, after the build, test the image in `--uses` with different level',
    )
    gp.add_argument(
        '--test-level',
        type=BuildTestLevel.from_string,
        choices=list(BuildTestLevel),
        default=BuildTestLevel.FLOW,
        help='If set, the test level when `--test-uses` is set, `NONE` means no test',
    )
    gp.add_argument(
        '--timeout-ready',
        type=int,
        default=600000,
        help='The timeout in millisecond to give for the Pod to start before considering a test failed',
    )
    gp.add_argument(
        '--host-info',
        action='store_true',
        default=False,
        help='If set, store the host information during bookkeeping',
    )
    gp.add_argument(
        '--daemon',
        action='store_true',
        default=False,
        help='If set, run the test Pea/Pod as a daemon process, see `jina pea --help` for details',
    )
    gp.add_argument(
        '--no-overwrite',
        action='store_true',
        default=False,
        help='If set, do not overwrite existing images (based on module version and jina version)',
    )



"""Module for helper functions for Docker."""

from pathlib import Path


def credentials_file() -> Path:
    """
    Get path to credentials file.

    :return: the path
    """
    Path.home().joinpath('.jina').mkdir(parents=True, exist_ok=True)
    return Path.home().joinpath('.jina').joinpath('access.yml')



def _update_autocomplete():
    from jina.parsers import get_main_parser

    def _gaa(key, parser):
        _result = {}
        _compl = []
        for v in parser._actions:
            if v.option_strings:
                _compl.extend(v.option_strings)
            elif v.choices:
                _compl.extend(v.choices)
                if isinstance(v.choices, dict):
                    for kk, vv in v.choices.items():
                        _result.update(_gaa(' '.join([key, kk]).strip(), vv))
        # filer out single dash, as they serve as abbrev
        _compl = [k for k in _compl if (not k.startswith('-') or k.startswith('--'))]
        _result.update({key: _compl})
        return _result

    compl = _gaa('', get_main_parser())
    cmd = compl.pop('')
    compl = {'commands': cmd, 'completions': compl}

    with open(__file__, 'a') as fp:
        fp.write(f'\nac_table = {compl}\n')


if __name__ == '__main__':
    _update_autocomplete()


ac_table = {
    'commands': [
        '--help',
        '--version',
        '--version-full',
        'hello',
        'executor',
        'pod',
        'flow',
        'ping',
        'gateway',
        'pea',
        'client',
        'export-api',
    ],
    'completions': {
        'hello fashion': [
            '--help',
            '--workdir',
            '--download-proxy',
            '--index-data-url',
            '--index-labels-url',
            '--query-data-url',
            '--query-labels-url',
            '--request-size',
            '--num-query',
            '--top-k',
        ],
        'hello chatbot': [
            '--help',
            '--workdir',
            '--download-proxy',
            '--index-data-url',
            '--port-expose',
            '--parallel',
            '--unblock-query-flow',
        ],
        'hello multimodal': [
            '--help',
            '--workdir',
            '--download-proxy',
            '--index-data-url',
            '--port-expose',
            '--unblock-query-flow',
        ],
        'hello fork': ['--help', 'fashion', 'chatbot', 'multimodal'],
        'hello': ['--help', 'fashion', 'chatbot', 'multimodal', 'fork'],
        'executor': [
            '--help',
            '--name',
            '--description',
            '--workspace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--num-part',
            '--freeze-network-settings',
            '--uses-internal',
            '--entrypoint',
            '--docker-kwargs',
            '--pull-latest',
            '--volumes',
            '--host',
            '--port-expose',
            '--proxy',
            '--quiet-remote-logs',
            '--upload-files',
            '--workspace-id',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--pea-id',
            '--pea-role',
            '--noblock-on-start',
            '--uses-before',
            '--uses-after',
            '--parallel',
            '--shards',
            '--replicas',
            '--polling',
            '--scheduling',
            '--external',
            '--pod-role',
            '--peas-hosts',
        ],
        'pod': [
            '--help',
            '--name',
            '--description',
            '--workspace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--num-part',
            '--freeze-network-settings',
            '--uses-internal',
            '--entrypoint',
            '--docker-kwargs',
            '--pull-latest',
            '--volumes',
            '--host',
            '--port-expose',
            '--proxy',
            '--quiet-remote-logs',
            '--upload-files',
            '--workspace-id',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--pea-id',
            '--pea-role',
            '--noblock-on-start',
            '--uses-before',
            '--uses-after',
            '--parallel',
            '--shards',
            '--replicas',
            '--polling',
            '--scheduling',
            '--external',
            '--pod-role',
            '--peas-hosts',
        ],
        'flow': [
            '--help',
            '--name',
            '--description',
            '--workspace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--uses',
            '--inspect',
        ],
        'ping': ['--help', '--timeout', '--retries', '--print-response'],
        'gateway': [
            '--help',
            '--name',
            '--description',
            '--workspace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--num-part',
            '--freeze-network-settings',
            '--max-message-size',
            '--prefetch',
            '--prefetch-on-recv',
            '--compress',
            '--compress-min-bytes',
            '--compress-min-ratio',
            '--including-default-value-fields',
            '--sort-keys',
            '--use-integers-for-enums',
            '--float-precision',
            '--host',
            '--port-expose',
            '--proxy',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--pea-id',
            '--pea-role',
            '--noblock-on-start',
            '--restful',
            '--rest-api',
        ],
        'pea': [
            '--help',
            '--name',
            '--description',
            '--workspace',
            '--log-config',
            '--quiet',
            '--quiet-error',
            '--identity',
            '--port-ctrl',
            '--ctrl-with-ipc',
            '--timeout-ctrl',
            '--ssh-server',
            '--ssh-keyfile',
            '--ssh-password',
            '--uses',
            '--py-modules',
            '--port-in',
            '--port-out',
            '--host-in',
            '--host-out',
            '--socket-in',
            '--socket-out',
            '--memory-hwm',
            '--on-error-strategy',
            '--num-part',
            '--freeze-network-settings',
            '--uses-internal',
            '--entrypoint',
            '--docker-kwargs',
            '--pull-latest',
            '--volumes',
            '--host',
            '--port-expose',
            '--proxy',
            '--quiet-remote-logs',
            '--upload-files',
            '--workspace-id',
            '--daemon',
            '--runtime-backend',
            '--runtime',
            '--runtime-cls',
            '--timeout-ready',
            '--env',
            '--expose-public',
            '--pea-id',
            '--pea-role',
            '--noblock-on-start',
        ],
        'client': [
            '--help',
            '--request-size',
            '--continue-on-error',
            '--show-progress',
            '--return-results',
            '--host',
            '--port-expose',
            '--proxy',
        ],
        'export-api': ['--help', '--yaml-path', '--json-path', '--schema-path'],
        'check': ['--help', '--summary-exec'],
    },
}



from ..base import set_base_parser


def set_hub_pushpull_parser(parser=None):
    """Set the parser for the hub push or hub pull

    :param parser: an optional existing parser to build upon
    :return: the parser
    """
    if not parser:
        parser = set_base_parser()

    from .login import mixin_hub_docker_login_parser
    from .pushpull import mixin_hub_pushpull_parser

    mixin_hub_docker_login_parser(parser)
    mixin_hub_pushpull_parser(parser)
    return parser


def set_hub_build_parser(parser=None):
    """Set the parser for `hub build`

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    from .login import mixin_hub_docker_login_parser
    from .build import mixin_hub_build_parser

    mixin_hub_docker_login_parser(parser)
    mixin_hub_build_parser(parser)

    return parser


def set_hub_list_parser(parser=None):
    """Set the parser for `hub list`

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    from .list import mixin_hub_list_parser

    mixin_hub_list_parser(parser)

    return parser


def set_hub_new_parser(parser=None):
    """Set the parser for the `hub new` command

    :param parser: the parser configure
    :return: the new parser
    """
    if not parser:
        parser = set_base_parser()

    from .new import mixin_hub_new_parser

    mixin_hub_new_parser(parser)

    return parser



"""Module wrapping the Client of Jina."""
import argparse
from typing import overload, Optional

from .base import BaseClient, CallbackFnType, InputType, InputDeleteType
from .helper import callback_exec
from .mixin import PostMixin
from .request import GeneratorSourceType
from .websocket import WebSocketClientMixin

__all__ = ['Client', 'GRPCClient', 'WebSocketClient']


# overload_inject_start_client
@overload
def Client(
    continue_on_error: Optional[bool] = False,
    host: Optional[str] = '0.0.0.0',
    port_expose: Optional[int] = None,
    proxy: Optional[bool] = False,
    request_size: Optional[int] = 100,
    restful: Optional[bool] = False,
    return_results: Optional[bool] = False,
    show_progress: Optional[bool] = False,
    **kwargs
) -> 'BaseClient':
    """Create a Client. Client is how user interact with Flow

    :param continue_on_error: If set, a Request that causes error will be logged only without blocking the further requests.
    :param host: The host address of the runtime, by default it is 0.0.0.0.
    :param port_expose: The port of the host exposed to the public
    :param proxy: If set, respect the http_proxy and https_proxy environment variables. otherwise, it will unset these proxy variables before start. gRPC seems to prefer no proxy
    :param request_size: The number of Documents in each Request.
    :param restful: If set, use RESTful interface instead of gRPC as the main interface
    :param return_results: This feature is only used for AsyncClient.

          If set, the results of all Requests will be returned as a list. This is useful when one wants
          process Responses in bulk instead of using callback.
    :param show_progress: If set, client will show a progress bar on receiving every request.
    :return: the new Client object

    .. # noqa: DAR202
    .. # noqa: DAR101
    .. # noqa: DAR003
    """
    # overload_inject_end_client


def Client(args: Optional['argparse.Namespace'] = None, **kwargs) -> 'BaseClient':
    """Jina Python client.

    :param args: Namespace args.
    :param kwargs: Additional arguments.
    :return: An instance of :class:`GRPCClient` or :class:`WebSocketClient`.
    """
    is_restful = kwargs.get('restful', False)
    if is_restful:
        return WebSocketClient(args, **kwargs)
    else:
        return GRPCClient(args, **kwargs)


class GRPCClient(PostMixin, BaseClient):
    """A simple Python client for connecting to the gRPC gateway.

    It manages the asyncio event loop internally, so all interfaces are synchronous from the outside.
    """

    @property
    def client(self) -> 'GRPCClient':
        """Return the client object itself
        .. # noqa: DAR201"""
        return self


class WebSocketClient(GRPCClient, WebSocketClientMixin):
    """A Python Client to stream requests from a Flow with a REST Gateway.

    :class:`WebSocketClient` shares the same interface as :class:`Client` and provides methods like
    :meth:`index`, "meth:`search`, :meth:`train`, :meth:`update` & :meth:`delete`.

    It is used by default while running operations when we create a `Flow` with `rest_api=True`

    .. highlight:: python
    .. code-block:: python

        from jina.flow import Flow
        f = Flow(rest_api=True).add().add()

        with f:
            f.index(['abc'])


    :class:`WebSocketClient` can also be used to run operations for a remote Flow

    .. highlight:: python
    .. code-block:: python

        # A Flow running on remote
        from jina.flow import Flow
        f = Flow(rest_api=True, port_expose=34567).add().add()

        with f:
            f.block()

        # Local WebSocketClient running index & search
        from jina.clients import WebSocketClient

        client = WebSocketClient(...)
        client.index(...)
        client.search(...)


    :class:`WebSocketClient` internally handles an event loop to run operations asynchronously.
    """
